#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `packmol_step` running full flowcharts."""
from pathlib import Path
import pytest
from seamm_exec import run

test_dir = Path(__file__).resolve().parent

flowcharts = [
    str(test_dir / path)
    for path in sorted(test_dir.glob("flowcharts/integration_*.flow"))
]
unit_flowcharts = [
    str(test_dir / path) for path in sorted(test_dir.glob("flowcharts/unit_*.flow"))
]


@pytest.mark.integration
@pytest.mark.parametrize("flowchart", flowcharts)
def test_ideal_gas(monkeypatch, tmp_path, flowchart):
    """Test creating a cell from an ideal gas."""
    monkeypatch.setattr(
        "sys.argv",
        [
            "testing",
            flowchart,
            "--standalone",
        ],
    )

    run(wdir=str(tmp_path))


@pytest.mark.unit
@pytest.mark.parametrize("flowchart", unit_flowcharts)
def test_unit(monkeypatch, tmp_path, flowchart):
    """Unit tests with flowcharts"""
    monkeypatch.setattr(
        "sys.argv",
        [
            "testing",
            flowchart,
            "--standalone",
        ],
    )

    run(wdir=str(tmp_path))



# -*- coding: utf-8 -*-

"""The graphical part of a Packmol step"""

import logging
import tkinter as tk
import tkinter.ttk as ttk

from packmol_step import PackmolParameters
import seamm
import seamm_widgets as sw

logger = logging.getLogger(__name__)


class TkPackmol(seamm.TkNode):
    """Graphical interface for using Packmol for fluid boxes"""

    def __init__(
        self, tk_flowchart=None, node=None, canvas=None, x=None, y=None, w=200, h=50
    ):
        """Initialize a node

        Keyword arguments:
        """

        self.dialog = None
        self._molecule_data = []
        self._add_molecule = None

        super().__init__(
            tk_flowchart=tk_flowchart, node=node, canvas=canvas, x=x, y=y, w=w, h=h
        )

    def create_dialog(self):
        """Create the dialog!"""
        frame = super().create_dialog(title="Edit PACKMOL parameters")

        # Create all the widgets
        P = self.node.parameters
        for key in P:
            if key not in ("molecules",):
                self[key] = P[key].widget(frame)

        # Frame for specifying molecules
        self["molecules"] = sw.ScrolledFrame(frame, height=500)
        w = self["molecules"].interior()
        self["component"] = ttk.Label(w, text="Component")
        self["source"] = ttk.Label(w, text="Source")
        self["definition"] = ttk.Label(w, text="Definition")
        self["stoichiometry"] = ttk.Label(w, text="proportion")

        # And the molecule data
        for molecule in P["molecules"].value:
            self._molecule_data.append({**molecule})

        for key in ("periodic", "shape", "dimensions", "fluid amount"):
            self[key].combobox.bind("<<ComboboxSelected>>", self.reset_dialog)
            self[key].combobox.bind("<Return>", self.reset_dialog)
            self[key].combobox.bind("<FocusOut>", self.reset_dialog)

        self.reset_dialog()

        # Resize the dialog to fill the screen, more or less.
        self.fit_dialog()

    def reset_dialog(self, widget=None):
        """Layout the widgets in the dialog contingent on the parameter values."""
        frame = self["frame"]

        # Reset everything
        columns, rows = frame.grid_size()
        for col in range(columns):
            frame.grid_columnconfigure(col, weight=0, minsize=0)
        for row in range(rows):
            frame.grid_rowconfigure(row, weight=0, minsize=0)
        for slave in frame.grid_slaves():
            slave.grid_forget()

        # The possible shapes depend on the periodicity
        periodic = self["periodic"].get()
        shape = self["shape"].get()
        if periodic == "Yes":
            shapes = PackmolParameters.periodic_shapes
        else:
            shapes = PackmolParameters.shapes
        self["shape"].combobox.config(values=shapes)
        if shape not in shapes:
            shape = shapes[0]
        self["shape"].set(shape)

        # The dimensions control the remaining widgets
        dimensions = self["dimensions"].get()

        widgets = []
        row = 0
        for key in ("periodic", "shape", "dimensions"):
            self[key].grid(row=row, column=0, sticky=tk.EW)
            row += 1
            widgets.append(self[key])

        if dimensions == "given explicitly":
            if shape == "cubic":
                keys = ("edge length",)
            elif shape == "rectangular":
                keys = ("a", "b", "c")
            elif shape == "spherical":
                keys = ("diameter",)
            else:
                raise RuntimeError(f"Do not recognize shape '{shape}'")
            for key in keys:
                self[key].grid(row=row, column=0, sticky=tk.EW)
                row += 1
                widgets.append(self[key])
        elif dimensions == "calculated from the volume":
            if shape == "rectangular":
                keys = ("volume", "a_ratio", "b_ratio", "c_ratio")
            else:
                keys = ("volume",)
            for key in keys:
                self[key].grid(row=row, column=0, sticky=tk.EW)
                row += 1
                widgets.append(self[key])
        elif dimensions == "calculated from the solute dimensions":
            keys = ("solvent thickness",)
            for key in keys:
                self[key].grid(row=row, column=0, sticky=tk.EW)
                row += 1
                widgets.append(self[key])
        elif dimensions == "calculated from the density":
            if shape == "rectangular":
                keys = ("density", "a_ratio", "b_ratio", "c_ratio")
            else:
                keys = ("density",)
            for key in keys:
                self[key].grid(row=row, column=0, sticky=tk.EW)
                row += 1
                widgets.append(self[key])
        elif dimensions == "calculated using the Ideal Gas Law":
            if shape == "rectangular":
                keys = ("temperature", "pressure", "a_ratio", "b_ratio", "c_ratio")
            else:
                keys = ("temperature", "pressure")
            for key in keys:
                self[key].grid(row=row, column=0, sticky=tk.EW)
                row += 1
                widgets.append(self[key])
        else:
            raise RuntimeError(f"Do not recognize dimensions '{dimensions}'")

        # Next we need the amount of material. However the options change depending on
        # the above
        amount = self["fluid amount"].get()
        if dimensions in (
            "calculated from the density",
            "calculated using the Ideal Gas Law",
        ):
            amounts = PackmolParameters.amounts_for_density
        elif dimensions in ("calculated from the solute dimensions",):
            amounts = PackmolParameters.amounts_for_layer
        else:
            amounts = PackmolParameters.amounts
        self["fluid amount"].combobox.config(values=amounts)
        if amount not in amounts:
            amount = amounts[0]
        self["fluid amount"].set(amount)

        for key in ("fluid amount",):
            self[key].grid(row=row, column=0, sticky=tk.EW)
            row += 1
            widgets.append(self[key])

        if amount == "rounding this number of atoms":
            for key in ("approximate number of atoms",):
                self[key].grid(row=row, column=0, sticky=tk.EW)
                row += 1
                widgets.append(self[key])
        elif amount == "rounding this number of molecules":
            for key in ("approximate number of molecules",):
                self[key].grid(row=row, column=0, sticky=tk.EW)
                row += 1
                widgets.append(self[key])
        elif amount == "using the density":
            for key in ("density",):
                self[key].grid(row=row, column=0, sticky=tk.EW)
                row += 1
                widgets.append(self[key])
        elif amount == "using the Ideal Gas Law":
            for key in ("temperature", "pressure"):
                self[key].grid(row=row, column=0, sticky=tk.EW)
                row += 1
                widgets.append(self[key])
        else:
            raise RuntimeError(f"Do not recognize amount '{amount}'")

        key = "assign forcefield"
        self[key].grid(row=row, column=0, sticky=tk.EW)
        row += 1
        widgets.append(self[key])

        sw.align_labels(widgets, sticky=tk.E)

        # The table of molecules to use
        self["molecules"].grid(row=row, column=0, columnspan=2, sticky=tk.NSEW)
        frame.grid_rowconfigure(row, weight=1, minsize=100)
        frame.grid_columnconfigure(1, weight=1)
        row += 1
        self.reset_molecules()

        # And finally, where to put the new system
        widgets = []
        for key in ("structure handling", "system name", "configuration name"):
            self[key].grid(row=row, column=0, sticky=tk.EW)
            row += 1
            widgets.append(self[key])

        sw.align_labels(widgets, sticky=tk.E)

    def reset_molecules(self):
        """Layout the table of molecules to use."""

        frame = self["molecules"].interior()

        # Unpack any widgets
        for slave in frame.grid_slaves():
            slave.grid_forget()

        # Put in the column headers.
        row = 0
        self["component"].grid(row=row, column=1, sticky=tk.EW)
        self["source"].grid(row=row, column=2, sticky=tk.EW)
        self["definition"].grid(row=row, column=3, sticky=tk.EW)
        self["stoichiometry"].grid(row=row, column=4, sticky=tk.EW)
        row += 1

        for data in self._molecule_data:
            if "widgets" not in data:
                widgets = data["widgets"] = {}
            else:
                widgets = data["widgets"]

            if "remove" not in widgets:
                # The button to remove a row...
                widgets["remove"] = ttk.Button(
                    frame,
                    text="-",
                    width=2,
                    command=lambda row=row: self.remove_molecule(row),
                    takefocus=True,
                )

            if "component" not in widgets:
                # the type of component
                widgets["component"] = ttk.Combobox(
                    frame,
                    values=("solute", "fluid"),
                    height=4,
                    width=6,
                    state="readonly",
                    takefocus=True,
                )
                widgets["component"].set(data["component"])

            if "source" not in widgets:
                # the type of source
                widgets["source"] = ttk.Combobox(
                    frame,
                    values=("configuration", "SMILES"),
                    height=4,
                    width=10,
                    state="readonly",
                    takefocus=True,
                )
                widgets["source"].set(data["source"])

            if "definition" not in widgets:
                # the SMILES ot system/configuration
                widgets["definition"] = ttk.Entry(frame, width=20, takefocus=True)
                widgets["definition"].insert("end", data["definition"])

            if "count" not in widgets:
                # The count for the stoichiometry
                widgets["count"] = ttk.Entry(frame, width=5, takefocus=True)
                widgets["count"].insert("end", data["count"])

            self.logger.debug("  widgets: " + str(widgets))
            widgets["remove"].grid(row=row, column=0, sticky=tk.W)
            widgets["component"].grid(row=row, column=1, stick=tk.EW)
            widgets["source"].grid(row=row, column=2, stick=tk.EW)
            widgets["definition"].grid(row=row, column=3, stick=tk.EW)
            widgets["count"].grid(row=row, column=4, stick=tk.EW)

            row += 1

        # The button to add a row...
        if self._add_molecule is None:
            self._add_molecule = ttk.Button(
                frame,
                text="+",
                width=5,
                command=self.add_molecule,
                takefocus=True,
            )
            self._add_molecule.focus_set()
        self._add_molecule.lift()
        self._add_molecule.grid(row=row, column=0, columnspan=3, sticky=tk.W)

        frame.grid_columnconfigure(3, weight=1)

    def right_click(self, event):
        """Probably need to add our dialog..."""

        super().right_click(event)
        self.popup_menu.add_command(label="Edit..", command=self.edit)

        self.popup_menu.tk_popup(event.x_root, event.y_root, 0)

    def handle_dialog(self, result):
        if result is None or result == "Cancel":
            self.dialog.deactivate(result)
            # Reset the molecules
            for data in self._molecule_data:
                if "widgets" in data:
                    widgets = data["widgets"]
                    for w in widgets.values():
                        w.destroy()
            self._molecule_data = []
            P = self.node.parameters
            for molecule in P["molecules"].value:
                self._molecule_data.append({**molecule})

            super().handle_dialog(result)

            return

        if result == "Help":
            # display help!!!
            return

        if result != "OK":
            self.dialog.deactivate(result)
            raise RuntimeError("Don't recognize dialog result '{}'".format(result))

        self.dialog.deactivate(result)

        # Shortcut for parameters
        P = self.node.parameters

        for key in P:
            if key not in ("molecules",):
                P[key].set_from_widget()

        # And handle the molecules
        molecules = []
        for data in self._molecule_data:
            widgets = data["widgets"]
            molecules.append(
                {
                    "component": widgets["component"].get(),
                    "source": widgets["source"].get(),
                    "definition": widgets["definition"].get(),
                    "count": widgets["count"].get(),
                }
            )
        P["molecules"].value = molecules

    def add_molecule(self):
        """Add a new row to the molecule table."""
        self._molecule_data.append(
            {
                "component": "fluid",
                "source": "SMILES",
                "definition": "",
                "count": "1",
            }
        )
        self.reset_molecules()

    def remove_molecule(self, row):
        """Remove a molecule entry from the table.

        Parameters
        ----------
        row : int
            The row in the table to remove. Note the first molecule is at row 1.
        """
        index = row - 1
        data = self._molecule_data[index]
        if "widgets" in data:
            for w in data["widgets"].values():
                w.destroy()
        del self._molecule_data[index]

        self.reset_molecules()



# -*- coding: utf-8 -*-

"""
packmol_step
A step for building fluid boxes using Packmol
"""

# Bring up the classes so that they appear to be directly in
# the packmol_step package.

from packmol_step.packmol import Packmol  # noqa: F401
from packmol_step.packmol_parameters import PackmolParameters  # noqa: F401
from packmol_step.packmol_step import PackmolStep  # noqa: F401
from packmol_step.tk_packmol import TkPackmol  # noqa: F401

# Handle versioneer
from ._version import get_versions

__author__ = """Paul Saxe"""
__email__ = "psaxe@molssi.org"
versions = get_versions()
__version__ = versions["version"]
__git_revision__ = versions["full-revisionid"]
del get_versions, versions



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Fixtures for testing the packmol_step package."""

import pytest

from molsystem import SystemDB


def pytest_addoption(parser):
    parser.addoption(
        "--no-unit", action="store_true", default=False, help="don't run the unit tests"
    )
    parser.addoption(
        "--integration",
        action="store_true",
        default=False,
        help="run the integration tests",
    )
    parser.addoption(
        "--timing", action="store_true", default=False, help="run the timing tests"
    )


def pytest_configure(config):
    config.addinivalue_line(
        "markers", "unit: unit tests run by default, use '--no-unit' to turn off"
    )
    config.addinivalue_line(
        "markers", "integration: integration test, run with --integration"
    )
    config.addinivalue_line("markers", "timing: timing tests, run with --timing")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--no-unit"):
        skip = pytest.mark.skip(reason="remove --no-unit option to run")
        for item in items:
            if "unit" in item.keywords:
                item.add_marker(skip)
    if not config.getoption("--integration"):
        skip = pytest.mark.skip(reason="use the --integration option to run")
        for item in items:
            if "integration" in item.keywords:
                item.add_marker(skip)
    if not config.getoption("--timing"):
        skip = pytest.mark.skip(reason="need --timing option to run")
        for item in items:
            if "timing" in item.keywords:
                item.add_marker(skip)


@pytest.fixture()
def empty_db():
    """Create a system db with no systems."""
    db = SystemDB(filename="file:seamm_db?mode=memory&cache=shared")

    yield db

    db.close()
    try:
        del db
    except:  # noqa: E722
        print("Caught error deleting the database")


@pytest.fixture()
def db(empty_db):
    """Create an empty system db."""
    system = empty_db.create_system(name="default")
    system.create_configuration(name="default")

    return empty_db


@pytest.fixture()
def system(db):
    """An empty system."""
    return db.system


@pytest.fixture()
def configuration(system):
    """An empty system."""
    return system.configuration


@pytest.fixture()
def Argon(configuration):
    """An system object for an argon atom"""
    configuration.name = "argon"
    configuration.atoms.append(x=0.0, y=0.0, z=0.0, symbol=["Ar"])

    return configuration



# -*- coding: utf-8 -*-

"""A step for building fluids with Packmol in a SEAMM flowchart"""

import configparser
import importlib
import logging
import math
from pathlib import Path
import pprint
import shutil
import textwrap

from tabulate import tabulate

from molsystem import SystemDB
import seamm
import seamm_util
from seamm_util import ureg, Q_, units_class  # noqa: F401
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __
import packmol_step

is_expr = seamm.Node.is_expr

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("packmol")


class Packmol(seamm.Node):
    def __init__(self, flowchart=None, extension=None):
        """Setup the main Packmol step

        Keyword arguments:
        """

        logger.debug("Handling arguments in Packmol {}".format(self))

        super().__init__(
            flowchart=flowchart,
            title="Packmol",
            extension=extension,
            module=__name__,
            logger=logger,
        )

        self.parameters = packmol_step.PackmolParameters()

    @property
    def version(self):
        """The semantic version of this module."""
        return packmol_step.__version__

    @property
    def git_revision(self):
        """The git version of this module."""
        return packmol_step.__git_revision__

    def create_parser(self):
        """Setup the command-line / config file parser"""
        parser_name = self.step_type
        parser = seamm_util.getParser()

        # Remember if the parser exists ... this type of step may have been
        # found before
        parser_exists = parser.exists(parser_name)

        # Create the standard options, e.g. log-level
        result = super().create_parser(name=parser_name)

        if parser_exists:
            return result

        return result

    def description_text(self, P=None):
        """Return a short description of this step.

        Return a nicely formatted string describing what this step will
        do.

        Keyword arguments:
            P: a dictionary of parameter values, which may be variables
                or final values. If None, then the parameters values will
                be used as is.
        """

        if not P:
            P = self.parameters.values_to_dict()

        periodic = P["periodic"]
        if isinstance(periodic, str):
            periodic = periodic.lower() == "yes"
        shape = P["shape"]

        if periodic:
            text = f"Will create a {shape} periodic cell"
        else:
            text = f"Will create a {shape} region"
        text += " containing the following molecules:\n\n"

        # Print table of molecules
        table = {
            "Component": [],
            "Structure": [],
            "Ratio": [],
        }
        for molecule in self.parameters["molecules"].value:
            table["Component"].append(molecule["component"])
            table["Structure"].append(molecule["definition"])
            if molecule["component"] == "solute":
                table["Ratio"].append("")
            else:
                table["Ratio"].append(molecule["count"])

        description = self.header + "\n" + str(__(text, indent=self.indent + 4 * " "))
        text_lines = tabulate(
            table, headers="keys", tablefmt="psql", colalign=("center", "left", "right")
        )
        description += textwrap.indent(text_lines, self.indent + 8 * " ")

        # And the rest of the control
        dimensions = P["dimensions"]
        text = f"\n\nThe dimensions of the region will be {dimensions}"
        if dimensions == "given explicitly":
            if shape == "cubic":
                text += f" by the edge length {P['edge length']}."
            elif shape == "rectangular":
                text += f" by the three sides {P['a']} x{P['b']} x{P['c']}."
            elif shape == "spherical":
                text += f" by the diameter {P['diameter']}."
            else:
                raise RuntimeError(f"Do not recognize shape '{shape}'")
        elif dimensions == "calculated from the volume":
            text += f" {P['volume']}."
        elif dimensions == "calculated from the solute dimensions":
            text += (
                ". If the input structure is periodic, its dimensions will be used. "
                "Otherwise for molecules, a the region will be a box with "
                f"extra space of {P['solvent thickness']} around the molecule."
            )
        elif dimensions == "calculated from the density":
            text += f" {P['density']}."
        elif dimensions == "calculated using the Ideal Gas Law":
            text += f" (PV=NRT) with P={P['pressure']} and T={P['temperature']}."
        else:
            raise RuntimeError(f"Do not recognize dimensions '{dimensions}'")

        amount = P["fluid amount"]
        text += " The number of molecules of the fluid will be obtained "
        if amount == "rounding this number of atoms":
            text += (
                f"by rounding {P['approximate number of atoms']} atoms to give "
                "a whole number of molecules with the requested ratios."
            )
        elif amount == "rounding this number of molecules":
            text += (
                f"by rounding {P['approximate number of molecules']} molecules to give "
                "the requested ratios of species."
            )
        elif amount == "using the density":
            text += f" by using the density {P['density']}."
        elif amount == "using the Ideal Gas Law":
            text += (
                f" by using the Ideal Gas Law (PV=NRT) with P={P['pressure']} and "
                f"T={P['temperature']}."
            )
        else:
            raise RuntimeError(f"Do not recognize amount '{amount}'")

        description += str(__(text, indent=self.indent + 4 * " "))
        return description

    def run(self):
        """Run a Packmol building step"""

        next_node = super().run(printer)

        # Get the system database
        system_db = self.get_variable("_system_db")

        # Access the options
        seamm_options = self.global_options

        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )
        periodic = P["periodic"]

        # Print what we are doing. Have to fix formatting for printing...
        PP = dict(P)
        for key in PP:
            if isinstance(PP[key], units_class):
                PP[key] = "{:~P}".format(PP[key])
        printer.important(self.description_text(PP))

        ff = None
        if P["assign forcefield"] != "No":
            try:
                ff = self.get_variable("_forcefield")
            except Exception:
                ff = None
            if ff == "OpenKIM":
                ff = None

        # Get the input files and any more output to print
        tmp_db = SystemDB(filename="file:tmp_db?mode=memory&cache=shared")
        molecules, files, output, cell = Packmol.get_input(
            P, system_db, tmp_db, seamm.flowchart_variables, ff=ff
        )

        self.logger.log(0, pprint.pformat(files))

        executor = self.flowchart.executor

        # Read configuration file for Packmol if it exists
        executor_type = executor.name
        full_config = configparser.ConfigParser()
        ini_dir = Path(seamm_options["root"]).expanduser()
        path = ini_dir / "packmol.ini"

        if path.exists():
            full_config.read(ini_dir / "packmol.ini")

        # If the section we need doesn't exists, get the default
        if not path.exists() or executor_type not in full_config:
            resources = importlib.resources.files("packmol_step") / "data"
            ini_text = (resources / "packmol.ini").read_text()
            full_config.read_string(ini_text)

        # Getting desperate! Look for an executable in the path
        if executor_type not in full_config:
            path = shutil.which("packmol")
            if path is None:
                raise RuntimeError(
                    f"No section for '{executor_type}' in Packmol ini file "
                    f"({ini_dir / 'packmol.ini'}), nor in the defaults, nor "
                    "in the path!"
                )
            else:
                full_config[executor_type] = {
                    "installation": "local",
                    "code": str(path),
                }

        # If the ini file does not exist, write it out!
        if not path.exists():
            with path.open("w") as fd:
                full_config.write(fd)
            printer.normal(f"Wrote the Packmol configuration file to {path}")
            printer.normal("")

        config = dict(full_config.items(executor_type))

        # Use the matching version of the seamm-packmol image by default.
        config["version"] = self.version

        result = executor.run(
            cmd=["{code}", "<", "input.inp", ">", "packmol.out"],
            config=config,
            directory=self.directory,
            files=files,
            return_files=["packmol.pdb"],
            in_situ=True,
            shell=True,
        )

        if not result:
            self.logger.error("There was an error running Packmol")
            return None

        self.logger.debug(pprint.pformat(result))

        # Get the bond orders and extra parameters like ff atom types
        bond_orders = []
        extra_data = {}
        total_q = 0.0
        for molecule in molecules:
            n = molecule["number"]
            orders = molecule["configuration"].bonds.get_column_data("bondorder")
            bond_orders.extend(orders * n)

            total_q += n * molecule["configuration"].charge

            atoms = molecule["configuration"].atoms
            for key in atoms.keys():
                if "atom_types_" in key or "charges" in key:
                    if key in extra_data:
                        extra_data[key].extend(atoms.get_column_data(key) * n)
                    else:
                        extra_data[key] = atoms.get_column_data(key) * n

        # Remove the temporary database
        tmp_db.close()

        # Get the system to fill and make sure it is empty
        system, configuration = self.get_system_configuration(P, same_as=None)
        configuration.clear()
        configuration.charge = total_q

        # Create the configuration from the PDB output of Packmol
        configuration.coordinate_system = "Cartesian"
        configuration.from_pdb_text(result["packmol.pdb"]["data"])

        # And set the bond orders and extra data we saved earlier.
        configuration.bonds.get_column("bondorder")[:] = bond_orders
        for key, values in extra_data.items():
            if key not in configuration.atoms:
                if "atom_types_" in key:
                    configuration.atoms.add_attribute(key, coltype="str")
                elif "charges" in key:
                    configuration.atoms.add_attribute(key, coltype="float")
                else:
                    raise RuntimeError(f"Can't handle extra column '{key}'")
            configuration.atoms.get_column(key)[:] = values

        # Finally, make periodic of correct size
        if periodic:
            configuration.periodicity = 3
            a, b, c = cell
            configuration.cell.parameters = (a, b, c, 90.0, 90.0, 90.0)
            # by convention we keep periodic systems in fractional coordinates
            xyz = configuration.atoms.get_coordinates(fractionals=False)
            configuration.coordinate_system = "fractional"
            configuration.atoms.set_coordinates(xyz, fractionals=False)

        printer.important(__(output, indent=4 * " "))
        printer.important("")

        # Since we have succeeded, add the citation.

        self.references.cite(
            raw=self._bibliography["doi:10.1002/jcc.21224"],
            alias="packmol",
            module="packmol_step",
            level=1,
            note="The principle Packmol citation.",
        )

        return next_node

    @staticmethod
    def get_input(P, system_db, tmp_db, context, ff=None):
        """Create the input for Packmol."""

        # Return the translation from points a to b
        def recenter(a, b):
            return b[0] - a[0], b[1] - a[1], b[2] - a[2]

        # Work out integer numbers of molecules, atoms, etc.
        def round_copies(n_copies, molecules):
            total_atoms = 0
            total_molecules = 0
            total_mass = 0.0
            total = 0.0
            total_count = 0.0
            for molecule in molecules:
                count = molecule["count"]
                component = molecule["type"]
                mass = molecule["mass"]
                n_atoms = molecule["n_atoms"]

                if component == "solute":
                    total_atoms += n_atoms
                    total_molecules += 1
                    total_mass += mass
                    molecule["number"] = 1
                else:
                    n = int(round(n_copies * count))
                    if n > 0:
                        total_atoms += n * n_atoms
                        total_molecules += n
                        total_mass += n * mass
                    molecule["number"] = n
                    total_count += molecule["count"]
                    total += n

            # Get the actual mol percent
            for molecule in molecules:
                if molecule["type"] == "solute":
                    molecule["actual %"] = ""
                    molecule["requested %"] = ""
                else:
                    molecule["actual %"] = f"{molecule['number']/total*100:.3f}"
                    molecule["requested %"] = f"{molecule['count']/total_count*100:.3f}"

            return total_atoms, total_molecules, total_mass

        # Start by handling the molecules
        n_solute_molecules = 0
        n_solute_atoms = 0
        solute_mass = 0.0
        n_fluid_molecules = 0
        n_fluid_atoms = 0
        fluid_mass = 0.0

        if ff is not None:
            ffname = ff.current_forcefield
            ff_key = f"atom_types_{ffname}"
            assign_ff_always = P["assign forcefield"] == "Always"

        # May need to create molecules.
        solute_configuration = None
        molecules = []
        for molecule in P["molecules"]:
            component = molecule["component"]
            if is_expr(component):
                component = context.value(component)
            source = molecule["source"]
            if is_expr(source):
                source = context.value(source)
            definition = molecule["definition"]
            if is_expr(definition):
                definition = context.value(definition)
            count = molecule["count"]
            if is_expr(count):
                count = context.value(count)
            count = float(count)

            if source == "SMILES":
                tmp_system = tmp_db.create_system(name=definition)
                tmp_configuration = tmp_system.create_configuration(name="default")
                tmp_configuration.from_smiles(definition, rdkit=True)
                if ff is not None:
                    ff.assign_forcefield(tmp_configuration)
            elif source == "configuration":
                if definition == "" or definition == "current":
                    tmp_system = system_db.system
                    tmp_configuration = tmp_system.configuration
                else:
                    if "/" in definition:
                        sysname, confname = definition.split("/")
                        tmp_system = system_db.get_system(sysname)
                    else:
                        confname = definition
                    tmp_configuration = tmp_system.get_configuration(confname)
                if ff is not None:
                    if ff_key not in tmp_configuration.atoms or assign_ff_always:
                        ff.assign_forcefield(tmp_configuration)

            tmp_mass = tmp_configuration.mass * ureg.g / ureg.mol
            tmp_mass.ito("kg")

            if component == "solute":
                n_solute_molecules += count
                n_solute_atoms += count * tmp_configuration.n_atoms
                solute_mass += count * tmp_mass
                if solute_configuration is None:
                    solute_configuration = tmp_configuration
                else:
                    raise RuntimeError("More than one solute system!")
            else:
                n_fluid_molecules += count
                n_fluid_atoms += count * tmp_configuration.n_atoms
                fluid_mass += count * tmp_mass

            molecules.append(
                {
                    "configuration": tmp_configuration,
                    "count": count,
                    "type": component,
                    "mass": tmp_mass,
                    "n_atoms": tmp_configuration.n_atoms,
                    "definition": definition,
                }
            )

        periodic = P["periodic"]
        shape = P["shape"]
        dimensions = P["dimensions"]
        amount = P["fluid amount"]
        cell = None

        # Get information about the solute for placing it
        if solute_configuration is not None:
            xyzs = solute_configuration.atoms.get_coordinates(fractionals=False)
            if shape == "cubic":
                center, sides = bounding_box(xyzs)
            elif shape == "rectangular":
                center, sides = bounding_box(xyzs)
            elif shape == "spherical":
                center, solute_radius = bounding_sphere(xyzs)

        # Work out the dimensions of the region
        if dimensions == "given explicitly":
            if shape == "cubic":
                a = P["edge length"].to("Å").magnitude
                if periodic:
                    cell = (a, a, a)
                    gap = P["gap"].to("Å").magnitude
                    a -= gap
                    x0 = f"{gap/2:.4f}"
                    region = f"   inside cube {x0} {x0} {x0} {a:.4f}"
                else:
                    region = f"   inside cube 0.0 0.0 0.0 {a:.4f}"
                if solute_configuration is not None:
                    dx, dy, dz = recenter(center, (a / 2, a / 2, a / 2))
                    fixed = f"   fixed {dx:.4f} {dy:.4f} {dz:.4f} 0.0 0.0 0.0"
                volume = a**3
            elif shape == "rectangular":
                a = P["a"].to("Å").magnitude
                b = P["b"].to("Å").magnitude
                c = P["c"].to("Å").magnitude
                if periodic:
                    cell = (a, b, c)
                    gap = P["gap"].to("Å").magnitude
                    a -= gap
                    b -= gap
                    c -= gap
                    x0 = f"{gap/2:.4f}"
                    region = f"   inside box {x0} {x0} {x0} {a:.4f} {b:.4f} {c:.4f}"
                else:
                    region = f"   inside box 0.0 0.0 0.0 {a:.4f} {b:.4f} {c:.4f}"
                if solute_configuration is not None:
                    dx, dy, dz = recenter(center, (a / 2, b / 2, c / 2))
                    fixed = f"   fixed {dx:.4f} {dy:.4f} {dz:.4f} 0.0 0.0 0.0"
                volume = a * b * c
            elif shape == "spherical":
                diameter = P["diameter"].to("Å").magnitude
                region = f"   inside sphere 0.0 0.0 0.0 {diameter/2:.4f}"
                if solute_configuration is not None:
                    dx, dy, dz = center
                    fixed = f"   fixed {-dx:.4f} {-dy:.4f} {-dz:.4f} 0.0 0.0 0.0"
                volume = 4 / 3 * math.pi * (diameter / 2) ** 3
            else:
                raise RuntimeError(f"Do not recognize shape '{shape}'")
        elif dimensions == "calculated from the volume":
            volume = P["volume"].to("Å^3").magnitude
            if shape == "cubic":
                a = volume ** (1 / 3)
                if periodic:
                    cell = (a, a, a)
                    gap = P["gap"].to("Å").magnitude
                    a -= gap
                    x0 = f"{gap/2:.4f}"
                    region = f"   inside cube {x0} {x0} {x0} {a:.4f}"
                else:
                    region = f"   inside cube 0.0 0.0 0.0 {a:.4f}"
                if solute_configuration is not None:
                    dx, dy, dz = recenter(center, (a / 2, a / 2, a / 2))
                    fixed = f"   fixed {dx:.4f} {dy:.4f} {dz:.4f} 0.0 0.0 0.0"
            elif shape == "rectangular":
                a = P["a"].to("Å").magnitude
                b = P["b"].to("Å").magnitude
                c = P["c"].to("Å").magnitude
                volume1 = a * b * c
                factor = (volume / volume1) ** (1 / 3)
                a *= factor
                b *= factor
                c *= factor
                if periodic:
                    cell = (a, b, c)
                    gap = P["gap"].to("Å").magnitude
                    a -= gap
                    b -= gap
                    c -= gap
                    x0 = f"{gap/2:.4f}"
                    region = f"   inside box {x0} {x0} {x0} {a:.4f} {b:.4f} {c:.4f}"
                else:
                    region = f"   inside box 0.0 0.0 0.0 {a:.4f} {b:.4f} {c:.4f}"
                if solute_configuration is not None:
                    dx, dy, dz = recenter(center, (a / 2, b / 2, c / 2))
                    fixed = f"   fixed {dx:.4f} {dy:.4f} {dz:.4f} 0.0 0.0 0.0"
            elif shape == "spherical":
                diameter = 2 * (volume / (4 / 3 * math.pi)) ** (1 / 3)
                region = f"   inside sphere 0.0 0.0 0.0 {diameter/2:.4f}"
                if solute_configuration is not None:
                    dx, dy, dz = center
                    fixed = f"   fixed {-dx:.4f} {-dy:.4f} {-dz:.4f} 0.0 0.0 0.0"
            else:
                raise RuntimeError(f"Do not recognize shape '{shape}'")
        elif dimensions == "calculated from the solute dimensions":
            if shape == "spherical":
                thickness = P["solvent thickness"].to("Å").magnitude
                region = f"   inside sphere 0.0 0.0 0.0 {solute_radius+thickness:.4f}"
                dx, dy, dz = center
                fixed = f"   fixed {-dx:.4f} {-dy:.4f} {-dz:.4f} 0.0 0.0 0.0"
                diameter = 2 * (solute_radius + thickness)
                volume = 4 / 3 * math.pi * (diameter / 2) ** 3
            else:
                thickness = P["solvent thickness"].to("Å").magnitude
                a, b, c = sides
                if shape == "cubic":
                    a = max(a, b, c)
                    if periodic:
                        volume = (a + thickness) ** 3
                        a += thickness
                        cell = (a, a, a)
                        gap = P["gap"].to("Å").magnitude
                        a -= gap
                        x0 = f"{gap/2:.4f}"
                        region = f"   inside cube {x0} {x0} {x0} {a:.4f}"
                    else:
                        a += 2 * thickness
                        volume = a**3
                        region = f"   inside cube 0.0 0.0 0.0 {a:.4f}"
                    # Move the solute molecule to the center of the box
                    dx, dy, dz = recenter(center, (a / 2, a / 2, a / 2))
                    fixed = f"   fixed {dx:.4f} {dy:.4f} {dz:.4f} 0.0 0.0 0.0"
                else:
                    if periodic:
                        a += thickness
                        b += thickness
                        c += thickness
                        cell = (a, b, c)
                        volume = a * b * c
                        gap = P["gap"].to("Å").magnitude
                        a -= gap
                        b -= gap
                        c -= gap
                        x0 = f"{gap/2:.4f}"
                        region = f"   inside box {x0} {x0} {x0} {a:.4f} {b:.4f} {c:.4f}"
                    else:
                        a += 2 * thickness
                        b += 2 * thickness
                        c += 2 * thickness
                        volume = a * b * c
                        region = f"   inside box 0.0 0.0 0.0 {a:.4f} {b:.4f} {c:.4f}"
                    # Move the solute molecule to the center of the box
                    dx, dy, dz = recenter(center, (a / 2, b / 2, c / 2))
                    fixed = f"   fixed {dx:.4f} {dy:.4f} {dz:.4f} 0.0 0.0 0.0"
        elif dimensions == "calculated from the density":
            density = P["density"]
            if amount == "rounding this number of atoms":
                n_atoms = P["approximate number of atoms"]
                n_copies = (n_atoms - n_solute_atoms) / n_fluid_atoms
            elif amount == "rounding this number of molecules":
                n_molecules = P["approximate number of molecules"]
                n_copies = n_molecules / n_fluid_molecules
            else:
                raise RuntimeError(f"Do not recognize fluid amount '{amount}'")
            n_atoms, n_molecules, mass = round_copies(n_copies, molecules)
            volume = mass / density
            volume = volume.to("Å^3").magnitude

            if shape == "cubic":
                a = volume ** (1 / 3)
                if solute_configuration is not None:
                    dx, dy, dz = recenter(center, (a / 2, a / 2, a / 2))
                    fixed = f"   fixed {dx:.4f} {dy:.4f} {dz:.4f} 0.0 0.0 0.0"
                if periodic:
                    cell = (a, a, a)
                    gap = P["gap"].to("Å").magnitude
                    a -= gap
                    x0 = f"{gap/2:.4f}"
                    region = f"   inside cube {x0} {x0} {x0} {a:.4f}"
                else:
                    region = f"   inside cube 0.0 0.0 0.0 {a:.4f}"
            elif shape == "rectangular":
                a = P["a"].to("Å").magnitude
                b = P["b"].to("Å").magnitude
                c = P["c"].to("Å").magnitude
                volume1 = a * b * c
                factor = (volume / volume1) ** (1 / 3)
                a *= factor
                b *= factor
                c *= factor
                if solute_configuration is not None:
                    dx, dy, dz = recenter(center, (a / 2, b / 2, c / 2))
                    fixed = f"   fixed {dx:.4f} {dy:.4f} {dz:.4f} 0.0 0.0 0.0"
                if periodic:
                    cell = (a, b, c)
                    gap = P["gap"].to("Å").magnitude
                    a -= gap
                    b -= gap
                    c -= gap
                    x0 = f"{gap/2:.4f}"
                    region = f"   inside box {x0} {x0} {x0} {a:.4f} {b:.4f} {c:.4f}"
                else:
                    region = f"   inside box 0.0 0.0 0.0 {a:.4f} {b:.4f} {c:.4f}"
            elif shape == "spherical":
                diameter = 2 * (volume / (4 / 3 * math.pi)) ** (1 / 3)
                region = f"   inside sphere 0.0 0.0 0.0 {diameter/2:.4f}"
                if solute_configuration is not None:
                    dx, dy, dz = center
                    fixed = f"   fixed {-dx:.4f} {-dy:.4f} {-dz:.4f} 0.0 0.0 0.0"
            else:
                raise RuntimeError(f"Do not recognize shape '{shape}'")
        elif dimensions == "calculated using the Ideal Gas Law":
            temperature = P["temperature"]
            pressure = P["pressure"]
            if amount == "rounding this number of atoms":
                n_atoms = P["approximate number of atoms"]
                n_copies = (n_atoms - n_solute_atoms) / n_fluid_atoms
            elif amount == "rounding this number of molecules":
                n_molecules = P["approximate number of molecules"]
                n_copies = n_molecules / n_fluid_molecules
            else:
                raise RuntimeError(f"Do not recognize fluid amount '{amount}'")
            n_atoms, n_molecules, mass = round_copies(n_copies, molecules)

            # PV = NRT
            volume = n_molecules * temperature * ureg.k / pressure
            volume = volume.to("Å^3").magnitude
            if shape == "cubic":
                a = volume ** (1 / 3)
                if solute_configuration is not None:
                    dx, dy, dz = recenter(center, (a / 2, a / 2, a / 2))
                    fixed = f"   fixed {dx:.4f} {dy:.4f} {dz:.4f} 0.0 0.0 0.0"
                if periodic:
                    cell = (a, a, a)
                    gap = P["gap"].to("Å").magnitude
                    a -= gap
                    x0 = f"{gap/2:.4f}"
                    region = f"   inside cube {x0} {x0} {x0} {a:.4f}"
                else:
                    region = f"   inside cube 0.0 0.0 0.0 {a:.4f}"
            elif shape == "rectangular":
                a = P["a"].to("Å").magnitude
                b = P["b"].to("Å").magnitude
                c = P["c"].to("Å").magnitude
                volume1 = a * b * c
                factor = (volume / volume1) ** (1 / 3)
                a *= factor
                b *= factor
                c *= factor
                if solute_configuration is not None:
                    dx, dy, dz = recenter(center, (a / 2, b / 2, c / 2))
                    fixed = f"   fixed {dx:.4f} {dy:.4f} {dz:.4f} 0.0 0.0 0.0"
                if periodic:
                    cell = (a, b, c)
                    gap = P["gap"].to("Å").magnitude
                    a -= gap
                    b -= gap
                    c -= gap
                    x0 = f"{gap/2:.4f}"
                    region = f"   inside box {x0} {x0} {x0} {a:.4f} {b:.4f} {c:.4f}"
                else:
                    region = f"   inside box 0.0 0.0 0.0 {a:.4f} {b:.4f} {c:.4f}"
            elif shape == "spherical":
                diameter = 2 * (volume / (4 / 3 * math.pi)) ** (1 / 3)
                region = f"   inside sphere 0.0 0.0 0.0 {diameter/2:.4f}"
                if solute_configuration is not None:
                    dx, dy, dz = center
                    fixed = f"   fixed {-dx:.4f} {-dy:.4f} {-dz:.4f} 0.0 0.0 0.0"
            else:
                raise RuntimeError(f"Do not recognize shape '{shape}'")
        else:
            raise RuntimeError(f"Do not recognize dimensions '{dimensions}'")

        # Now that we have the volume, get the number of molecules
        if amount == "rounding this number of atoms":
            n_atoms = P["approximate number of atoms"]
            n_copies = (n_atoms - n_solute_atoms) / n_fluid_atoms
        elif amount == "rounding this number of molecules":
            n_molecules = P["approximate number of molecules"]
            n_copies = n_molecules / n_fluid_molecules
        elif amount == "using the density":
            density = P["density"]
            mass = Q_(volume, "Å^3") * density
            mass.ito("kg")
            n_copies = (mass - solute_mass) / fluid_mass
        elif amount == "using the Ideal Gas Law":
            # PV = NRT
            n_molecules = pressure * volume / (temperature * ureg.k)
            n_copies = n_molecules / n_fluid_molecules
        else:
            raise RuntimeError(f"Do not recognize fluid amount '{amount}'")
        n_atoms, n_molecules, mass = round_copies(n_copies, molecules)

        # Prepare the input
        lines = []
        lines.append("tolerance 2.0")
        lines.append("output packmol.pdb")
        lines.append("filetype pdb")
        lines.append("connect yes")

        files = {}
        for i, molecule in enumerate(molecules, start=1):
            lines.append(f"structure input_{i}.pdb")
            lines.append(region)
            if molecule["type"] == "solute":
                lines.append("   center")
                lines.append(fixed)
                lines.append("   number 1")
            else:
                lines.append(f"   number {molecule['number']}")
            lines.append("end structure")
            configuration = molecule["configuration"]
            files[f"input_{i}.pdb"] = configuration.to_pdb_text()

        lines.append("")
        files["input.inp"] = "\n".join(lines)

        string = "\n"
        if periodic:
            a, b, c = cell
            if shape == "cubic":
                string += f"Created a periodic cubic cell {a:.2f} Å on a side"
            else:
                string += f"Created a periodic {a:.2f} x {b:.2f} x {c:.2f} Å cell"
        else:
            if shape == "cubic":
                string += f"Created a cubic region {a:.2f} Å on a side"
            elif shape == "rectangular":
                string += f"Created a rectangular {a:.2f} x {b:.2f} x {c:.2f} Å region"
            else:
                string += (
                    f"Created a spherical region with a diameter of {diameter:.2f} Å"
                )
        volume = Q_(volume, "Å^3")
        density = mass / volume
        density.ito("g/ml")

        if n_solute_molecules > 0:
            string += (
                f" with the solute and {n_molecules-n_solute_molecules} solvent "
                "molecules\n\n"
            )
        else:
            string += f" with {n_molecules} fluid molecules\n\n"

        # Reprint table of molecules
        table = {
            "Component": [],
            "Structure": [],
            "Requested %": [],
            "Number": [],
            "Actual %": [],
        }
        for molecule in molecules:
            table["Component"].append(molecule["type"])
            table["Structure"].append(molecule["definition"])
            table["Requested %"].append(molecule["requested %"])
            table["Number"].append(molecule["number"])
            table["Actual %"].append(molecule["actual %"])

        text_lines = tabulate(
            table, headers="keys", tablefmt="psql", colalign=("center", "left", "right")
        )
        string += textwrap.indent(text_lines, 4 * " ")

        string += f"\n\nThere are a total of {n_atoms} atoms in the cell"
        string += f" giving a density of {density:.5~P}."

        return molecules, files, string, cell


def bounding_sphere(points):
    """A fast, approximate method for finding the sphere containing a set of points.

    See https://www.researchgate.net/publication/242453691_An_Efficient_Bounding_Sphere

    This method is approximate. While the sphere is guaranteed to contain all the points
    it is a few percent larger than necessary on average.
    """

    # euclidean metric
    def dist(a, b):
        return ((a[0] - b[0]) ** 2 + (a[1] - b[1]) ** 2 + (a[2] - b[2]) ** 2) ** 0.5

    def cent(a, b):
        return ((a[0] + b[0]) / 2, (a[1] + b[1]) / 2, (a[2] + b[2]) / 2)

    p0 = points[0]  # any arbitrary point in the point cloud works
    # choose point y furthest away from x
    p1 = max(points, key=lambda p: dist(p, p0))
    # choose point z furthest away from y
    p2 = max(points, key=lambda p: dist(p, p1))

    # initial bounding sphere
    center = cent(p1, p2)
    radius = dist(p1, p2) / 2

    # while there are points lying outside the bounding sphere, update the sphere by
    # growing it to fit
    for p in points:
        distance = dist(p, center)
        if distance > radius:
            delta = (distance - radius) / 2
            radius = (radius + distance) / 2

            cx, cy, cz = center
            x, y, z = p
            cx += (x - cx) / distance * delta
            cy += (y - cy) / distance * delta
            cz += (z - cz) / distance * delta

            center = (cx, cy, cz)

    return (center, radius)


def bounding_box(points):
    minx, miny, minz = points[0]
    maxx, maxy, maxz = points[0]
    for p in points:
        x, y, z = p
        minx = x if x < minx else minx
        miny = y if y < miny else miny
        minz = z if z < minz else minz
        maxx = x if x > maxx else maxx
        maxy = y if y > maxy else maxy
        maxz = z if z > maxz else maxz

    return (
        ((minx + maxx) / 2, (miny + maxy) / 2, (minz + maxz) / 2),
        ((maxx - minx), (maxy - miny), (maxz - minz)),
    )



# !/usr/bin/env python
# -*- coding: utf-8 -*-

"""Handle the installation of the PACKMOL step."""

from .installer import Installer


def run():
    """Handle the extra installation needed.

    * Find and/or install the PACKMOL executables.
    * Add or update information in the SEAMM.ini file for PACKMOL
    """

    # print('The is the installer for the PACKMOL step.')
    # Create an installer object
    installer = Installer()
    installer.run()


if __name__ == "__main__":
    run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `packmol_step` package."""

import json
from pathlib import Path
import seamm
from seamm_util import units_class

import pytest

from molsystem import SystemDB
import packmol_step

test_dir = Path(__file__).resolve().parent
inputs = [test_dir / path for path in sorted(test_dir.glob("inputs/*.json"))]


@pytest.mark.unit
@pytest.mark.parametrize("input_file", inputs)
def test_spherical_region(input_file):
    """Test the PACKMOL input and description"""

    # The parameters for the test
    with open(input_file) as fd:
        tmp = json.load(fd)

    # Handle any local variables that need definition
    variables = tmp.pop("_variables_", {})
    seamm.flowchart_variables = seamm.Variables(**variables)

    parameters = packmol_step.PackmolParameters()
    parameters.from_dict(tmp)
    P = parameters.current_values_to_dict(context=seamm.flowchart_variables._data)
    # The golden output file and PACKMOL input file
    output = test_dir / "outputs" / input_file.with_suffix(".out").name
    input = test_dir / "outputs" / input_file.with_suffix(".inp").name

    # Create the two System Databases that are needed
    system_db = SystemDB(filename="file:seamm_db?mode=memory&cache=shared")
    tmp_db = SystemDB(filename="file:tmp_db?mode=memory&cache=shared")

    # Print what we are doing. Have to fix formatting for printing...
    PP = dict(P)
    for key in PP:
        if isinstance(PP[key], units_class):
            PP[key] = "{:~P}".format(PP[key])

    instance = packmol_step.Packmol()
    instance._id = (1,)
    instance.parameters = parameters

    # Get the description
    description = instance.description_text(PP)
    output_text = "\n".join(description.splitlines()[1:])

    # And get the input for PACKMOL
    molecules, files, text, cell = packmol_step.Packmol.get_input(
        P, system_db, tmp_db, seamm.flowchart_variables
    )

    output_text += "\n"
    output_text += text

    output_reference = ""
    input_reference = ""
    if output.exists():
        output_reference = output.read_text()
        if output_text != output_reference:
            new_output = output.with_suffix(".out_new")
            new_output.write_text(output_text)
            print(f"Updated output written to {new_output}")
    else:
        output.write_text(output_text)
        print(f"New output written to {output}")

    if input.exists():
        input_reference = input.read_text()
        if files["input.inp"] != input_reference:
            new_input = input.with_suffix(".inp_new")
            new_input.write_text(files["input.inp"])
            print(f"Updated PACKMOL input written to {new_input}")
    else:
        input.write_text(files["input.inp"])
        print(f"New PACKMOL input written to {input}")

    assert output_text == output_reference
    assert files["input.inp"] == input_reference



# -*- coding: utf-8 -*-

"""Installer for the PACKMOL plug-in.

This handles any further installation needed after installing the Python
package `packmol-step`.
"""

import logging
from pathlib import Path
import pkg_resources
import subprocess

import seamm_installer

logger = logging.getLogger(__name__)


class Installer(seamm_installer.InstallerBase):
    """Handle further installation needed after installing packmol-step.

    The Python package `packmol-step` should already be installed, using `pip`,
    `conda`, or similar. This plug-in-specific installer then checks for the
    PACKMOL executable, installing it if needed, and registers its
    location in seamm.ini.

    There are a number of ways to determine which are the correct PACKMOL
    executables to use. The aim of this installer is to help the user locate
    the executables. There are a number of possibilities

    1.  The correct executables are already available.

        1. If they are already registered in `seamm.ini` there is nothing else
           to do.
        #. They may be in the current path, in which case they need to be added
           to `seamm.ini`.
        #. If a module system is in use, a module may need to be loaded to give
           access to PACKMOL.
        #. They cannot be found automatically, so the user needs to locate the
           executables for the installer.

    #.  PACKMOL is not installed on the machine. In this case they can be
        installed in a Conda environment. There is one choice

        1. They can be installed in a separate environment, `seamm-packmol` by
           default.
    """

    def __init__(self, logger=logger):
        # Call the base class initialization, which sets up the commandline
        # parser, amongst other things.
        super().__init__(logger=logger)

        logger.debug("Initializing the PACKMOL installer object.")

        self.section = "packmol-step"
        self.executables = ["packmol"]
        self.resource_path = Path(pkg_resources.resource_filename(__name__, "data/"))

        # The environment.yaml file for Conda installations.
        logger.debug(f"data directory: {self.resource_path}")
        self.environment_file = self.resource_path / "seamm-packmol.yml"

    def check(self):
        """Check the status of the Packmol installation."""
        print("Checking the Packmol installation.")

        # What Conda environment is the default?
        path = self.configuration.path.parent / "packmol.ini"
        if not path.exists():
            text = (self.resource_path / "packmol.ini").read_text()
            path.write_text(text)
            print(f"    The packmol.ini file did not exist. Created {path}")

        self.exe_config.path = path

        # Get the current values
        data = self.exe_config.get_values("local")

        if "conda-environment" in data and data["conda-environment"] != "":
            self.environment = data["conda-environment"]
        else:
            self.environment = "seamm-packmol"

        super().check()

    def install(self):
        """Install Packmol in a conda environment."""
        print("Installing Packmol.")

        # What Conda environment is the default?
        path = self.configuration.path.parent / "packmol.ini"
        if not path.exists():
            text = (self.resource_path / "packmol.ini").read_text()
            path.write_text(text)
            print(f"    The packmol.ini file did not exist. Created {path}")

        self.exe_config.path = path

        # Get the current values
        data = self.exe_config.get_values("local")

        if "conda-environment" in data and data["conda-environment"] != "":
            self.environment = data["conda-environment"]
        else:
            self.environment = "seamm-packmol"

        super().install()

    def show(self):
        """Show the status of the Packmol installation."""
        print("Showing the Packmol installation.")

        # What Conda environment is the default?
        path = self.configuration.path.parent / "packmol.ini"
        if not path.exists():
            text = (self.resource_path / "packmol.ini").read_text()
            path.write_text(text)
            print(f"    The packmol.ini file does not exist at {path}")
            print("    The 'check' command will create it if Packmol is installed.")
            print("    Otherwise 'install' will install Packmol.")
            return

        self.exe_config.path = path

        if not self.exe_config.section_exists("local"):
            print(
                "    Packmol is not configured: there is no 'local' section in "
                f"     {path}."
            )
            return

        # Get the current values
        data = self.exe_config.get_values("local")

        if "conda-environment" in data and data["conda-environment"] != "":
            self.environment = data["conda-environment"]
        else:
            self.environment = "seamm-packmol"

        super().show()

    def uninstall(self):
        """Uninstall the Packmol installation."""
        print("Uninstall the Packmol installation.")

        # What Conda environment is the default?
        path = self.configuration.path.parent / "packmol.ini"
        if not path.exists():
            text = (self.resource_path / "packmol.ini").read_text()
            path.write_text(text)
            print(
                f""""    The packmol.ini file does not exist at {path}
    Perhaps Packmol is not installed, but if it is the 'check' command may locate it
    and create the ini file, after which 'uninstall' will remove it."""
            )
            return

        self.exe_config.path = path

        if not self.exe_config.section_exists("local"):
            print(
                f""""    The packmol.ini file at {path} does not have local section.
    Perhaps Packmol is not installed, but if it is the 'check' command may locate it
    and update the ini file, after which 'uninstall' will remove it."""
            )
            return

        # Get the current values
        data = self.exe_config.get_values("local")

        if "conda-environment" in data and data["conda-environment"] != "":
            self.environment = data["conda-environment"]
        else:
            self.environment = "seamm-packmol"

        super().uninstall()

    def update(self):
        """Updates the Packmol installation."""
        print("Updating the Packmol installation.")

        # What Conda environment is the default?
        path = self.configuration.path.parent / "packmol.ini"
        if not path.exists():
            text = (self.resource_path / "packmol.ini").read_text()
            path.write_text(text)
            print(f"    The packmol.ini file did not exist. Created {path}")

        self.exe_config.path = path

        # Get the current values
        data = self.exe_config.get_values("local")

        if "conda-environment" in data and data["conda-environment"] != "":
            self.environment = data["conda-environment"]
        else:
            self.environment = "seamm-packmol"

        super().update()

    def exe_version(self, config):
        """Get the version of the PACKMOL executable.

        Parameters
        ----------
        config : dict
            Dictionary of options for running Packmol

        Returns
        -------
        "Packmol", str
            The version reported by the executable, or 'unknown'.
        """
        environment = config["conda-environment"]
        conda = config["conda"]
        if environment[0] == "~":
            environment = str(Path(environment).expanduser())
            command = f"'{conda}' run --live-stream -p '{environment}'"
        elif Path(environment).is_absolute():
            command = f"'{conda}' run --live-stream -p '{environment}'"
        else:
            command = f"'{conda}' run --live-stream -n '{environment}'"
        command += " packmol -log none"

        logger.debug(f"    Running {command}")
        try:
            result = subprocess.run(
                command,
                stdin=subprocess.DEVNULL,
                capture_output=True,
                text=True,
                shell=True,
            )
        except Exception:
            version = "unknown"
        else:
            version = "unknown"
            lines = result.stdout.splitlines()
            for line in lines:
                line = line.strip()
                tmp = line.split()
                if len(tmp) == 2:
                    key, value = tmp
                    if key == "Version":
                        version = value
                        break

        return "Packmol", version



# -*- coding: utf-8 -*-
"""Control parameters for Packmol, currently for packing fluids
"""

import logging
import seamm

logger = logging.getLogger(__name__)


class PackmolParameters(seamm.Parameters):
    """The control parameters for Packmol packing fluids"""

    shapes = (
        "cubic",
        "rectangular",
        "spherical",
    )
    periodic_shapes = (
        "cubic",
        "rectangular",
    )
    amounts = (
        "rounding this number of atoms",
        "rounding this number of molecules",
        "using the density",
        "using the Ideal Gas Law",
    )
    amounts_for_layer = (
        "using the density",
        "using the Ideal Gas Law",
    )
    amounts_for_density = (
        "rounding this number of atoms",
        "rounding this number of molecules",
    )

    parameters = {
        "molecules": {
            "default": [],
            "kind": "list",
            "default_units": None,
            "enumeration": tuple(),
            "format_string": "",
            "description": "The molecules",
            "help_text": "An internal place to put the molecule definitions.",
        },
        "periodic": {
            "default": "No",
            "kind": "boolean",
            "default_units": "",
            "enumeration": (
                "Yes",
                "No",
            ),
            "format_string": "s",
            "description": "Create periodic system:",
            "help_text": "Whether to create a periodic system or not.",
        },
        "shape": {
            "default": "spherical",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": shapes,
            "format_string": "s",
            "description": "Shape of the region:",
            "help_text": "The shape of the desired region.",
        },
        "dimensions": {
            "default": "calculated from the density",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": (
                "given explicitly",
                "calculated from the volume",
                "calculated from the solute dimensions",
                "calculated from the density",
                "calculated using the Ideal Gas Law",
            ),
            "format_string": "s",
            "description": "The dimensions will be",
            "help_text": "How to get the dimensions of the region",
        },
        "fluid amount": {
            "default": "rounding this number of atoms",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": amounts_for_density,
            "format_string": "s",
            "description": "Get number of fluid molecules by",
            "help_text": "How to get the number of fluid molecules",
        },
        "density": {
            "default": 0.7,
            "kind": "float",
            "default_units": "g/ml",
            "enumeration": tuple(),
            "format_string": ".1f",
            "description": "Density:",
            "help_text": ("The target density of the cell."),
        },
        "volume": {
            "default": 8.0,
            "kind": "float",
            "default_units": "nm^3",
            "enumeration": tuple(),
            "format_string": ".1f",
            "description": "The volume of the cell:",
            "help_text": ("The volume of the target cell."),
        },
        "temperature": {
            "default": 298.15,
            "kind": "float",
            "default_units": "K",
            "enumeration": tuple(),
            "format_string": ".2f",
            "description": "T:",
            "help_text": ("The temperature using an ideal gas model (PV=nRT)."),
        },
        "pressure": {
            "default": 1.0,
            "kind": "float",
            "default_units": "atm",
            "enumeration": tuple(),
            "format_string": ".2f",
            "description": "P:",
            "help_text": ("The pressure using an ideal gas model (PV=nRT)."),
        },
        "gap": {
            "default": 2.0,
            "kind": "float",
            "default_units": "Å",
            "enumeration": tuple(),
            "format_string": ".1f",
            "description": "Gap around cell:",
            "help_text": (
                "Since Packmol does not support periodic systems "
                "we will build a box with this gap around the "
                "atoms, then make it periodic. The gap ensures "
                "that molecules at the boundary do not hit images"
            ),
        },
        "edge length": {
            "default": 20,
            "kind": "float",
            "default_units": "Å",
            "enumeration": tuple(),
            "format_string": ".1f",
            "description": "Length of the cube edge:",
            "help_text": ("The length of the cube edge."),
        },
        "a": {
            "default": 20,
            "kind": "float",
            "default_units": "Å",
            "enumeration": tuple(),
            "format_string": ".1f",
            "description": "a:",
            "help_text": "The length of the first side of the box.",
        },
        "b": {
            "default": 20,
            "kind": "float",
            "default_units": "Å",
            "enumeration": tuple(),
            "format_string": ".1f",
            "description": "b:",
            "help_text": "The length of the second side of the box.",
        },
        "c": {
            "default": 20,
            "kind": "float",
            "default_units": "Å",
            "enumeration": tuple(),
            "format_string": ".1f",
            "description": "c:",
            "help_text": "The length of the third side of the box.",
        },
        "a_ratio": {
            "default": 1,
            "kind": "float",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": ".1f",
            "description": "a ratio:",
            "help_text": "The ratio for the first side of the box.",
        },
        "b_ratio": {
            "default": 1,
            "kind": "float",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": ".1f",
            "description": "b ratio:",
            "help_text": "The ratio for the second side of the box.",
        },
        "c_ratio": {
            "default": 1,
            "kind": "float",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": ".1f",
            "description": "c ratio:",
            "help_text": "The ratio for the third side of the box.",
        },
        "diameter": {
            "default": 20.0,
            "kind": "float",
            "default_units": "Å",
            "enumeration": tuple(),
            "format_string": ".1f",
            "description": "The diameter:",
            "help_text": "The diameter of the sphere or cylinder.",
        },
        "solvent thickness": {
            "default": 10.0,
            "kind": "float",
            "default_units": "Å",
            "enumeration": tuple(),
            "format_string": ".1f",
            "description": "Solvent thickness:",
            "help_text": "The thickness of the layer of solvent around the solute",
        },
        "approximate number of molecules": {
            "default": 100,
            "kind": "integer",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": "d",
            "description": "Approximate number of molecules:",
            "help_text": ("The approximate number of molecules to pack in the cell."),
        },
        "approximate number of atoms": {
            "default": 2000,
            "kind": "integer",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": "d",
            "description": "Approximate number of atoms:",
            "help_text": (
                "The approximate number of atoms packed into the "
                "cell. This will be rounded to give whole molecules"
            ),
        },
        "assign forcefield": {
            "default": "If not assigned",
            "kind": "enum",
            "default_units": "",
            "enumeration": (
                "If not assigned",
                "Always",
                "No",
            ),
            "format_string": "s",
            "description": "Assign forcefield:",
            "help_text": "Whether to assign the forcefield to the molecules.",
        },
    }

    def __init__(self, defaults={}, data=None):
        """Initialize the instance, by default from the default
        parameters given in the class"""

        super().__init__(
            defaults={
                **PackmolParameters.parameters,
                **seamm.standard_parameters.structure_handling_parameters,
                **defaults,
            },
            data=data,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""packmol_step
A SEAMM plug-in for building periodic boxes of fluid using Packmol
"""

import sys
from setuptools import setup, find_packages
import versioneer

short_description = __doc__.splitlines()[1]

# from https://github.com/pytest-dev/pytest-runner#conditional-requirement
needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
pytest_runner = ['pytest-runner'] if needs_pytest else []

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

with open('requirements_install.txt') as fd:
    requirements = fd.read()

setup(
    # Descriptive entries which should always be present
    name='packmol_step',
    author="Paul Saxe",
    author_email='psaxe@molssi.org',
    description=short_description,
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/x-rst',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    license='BSD-3-Clause',
    url='https://github.com/molssi-seam/packmol_step',
    packages=find_packages(include=['packmol_step']),
    include_package_data=True,
    setup_requires=[] + pytest_runner,
    install_requires=requirements,
    test_suite='tests',
    platforms=['Linux',
               'Mac OS-X',
               'Unix',
               'Windows'],

    # Manual control if final package is compressible or not, set False to
    # prevent the .egg from being made
    zip_safe=True,

    keywords=['SEAMM', 'plug-in', 'flowchart', 'Packmol', 'fluid',
              'molecules'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    entry_points={
        'console_scripts': [
            'packmol-step-installer=packmol_step.__main__:run',
        ],
        'org.molssi.seamm': [
            'Packmol = packmol_step:PackmolStep',
        ],
        'org.molssi.seamm.tk': [
            'Packmol = packmol_step:PackmolStep',
        ],
    }
)
